@@ -6,7 +6,7 @@
 **
 ** This file is part of the QtWaylandCompositor module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:GPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
@@ -15,24 +15,14 @@
 ** and conditions see https://www.qt.io/terms-conditions. For further
 ** information use the contact form at https://www.qt.io/contact-us.
 **
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** General Public License version 3 or (at your option) any later version
+** approved by the KDE Free Qt Foundation. The licenses are as published by
+** the Free Software Foundation and appearing in the file LICENSE.GPL3
 ** included in the packaging of this file. Please review the following
 ** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -49,7 +39,6 @@
 #include "wayland_wrapper/qwldatadevicemanager_p.h"
 #endif
 
-#include "extensions/qwlextendedsurface_p.h"
 #include "qwaylandinputmethodcontrol_p.h"
 
 #include <QtWaylandCompositor/QWaylandCompositor>
@@ -60,6 +49,7 @@
 #include <QtWaylandCompositor/private/qwaylandcompositor_p.h>
 #include <QtWaylandCompositor/private/qwaylandview_p.h>
 #include <QtWaylandCompositor/private/qwaylandseat_p.h>
+#include <QtWaylandCompositor/private/qwaylandutils_p.h>
 
 #include <QtCore/private/qobject_p.h>
 
@@ -67,6 +57,7 @@
 #include <QtGui/QScreen>
 
 #include <QtCore/QDebug>
+#include <QtCore/QtMath>
 
 QT_BEGIN_NAMESPACE
 
@@ -76,14 +67,8 @@ public:
     FrameCallback(QWaylandSurface *surf, wl_resource *res)
         : surface(surf)
         , resource(res)
-        , canSend(false)
     {
-#if WAYLAND_VERSION_MAJOR < 1 || (WAYLAND_VERSION_MAJOR == 1 && WAYLAND_VERSION_MINOR <= 2)
-        res->data = this;
-        res->destroy = destroyCallback;
-#else
-        wl_resource_set_implementation(res, 0, this, destroyCallback);
-#endif
+        wl_resource_set_implementation(res, nullptr, this, destroyCallback);
     }
     ~FrameCallback()
     {
@@ -102,18 +87,14 @@ public:
     }
     static void destroyCallback(wl_resource *res)
     {
-#if WAYLAND_VERSION_MAJOR < 1 || (WAYLAND_VERSION_MAJOR == 1 && WAYLAND_VERSION_MINOR <= 2)
-        FrameCallback *_this = static_cast<FrameCallback *>(res->data);
-#else
         FrameCallback *_this = static_cast<FrameCallback *>(wl_resource_get_user_data(res));
-#endif
         if (_this->surface)
             QWaylandSurfacePrivate::get(_this->surface)->removeFrameCallback(_this);
         delete _this;
     }
-    QWaylandSurface *surface;
-    wl_resource *resource;
-    bool canSend;
+    QWaylandSurface *surface = nullptr;
+    wl_resource *resource = nullptr;
+    bool canSend = false;
 };
 }
 static QRegion infiniteRegion() {
@@ -126,22 +107,7 @@ QList<QWaylandSurfacePrivate *> QWaylandSurfacePrivate::uninitializedSurfaces;
 #endif
 
 QWaylandSurfacePrivate::QWaylandSurfacePrivate()
-    : QtWaylandServer::wl_surface()
-    , compositor(Q_NULLPTR)
-    , refCount(1)
-    , client(Q_NULLPTR)
-    , role(0)
-    , inputRegion(infiniteRegion())
-    , bufferScale(1)
-    , isCursorSurface(false)
-    , destroyed(false)
-    , hasContent(false)
-    , isInitialized(false)
-    , contentOrientation(Qt::PrimaryOrientation)
-#if QT_CONFIG(im)
-    , inputMethodControl(Q_NULLPTR)
-#endif
-    , subsurface(0)
+    : inputRegion(infiniteRegion())
 {
     pending.buffer = QWaylandBufferRef();
     pending.newlyAttached = false;
@@ -161,31 +127,12 @@ QWaylandSurfacePrivate::~QWaylandSurfacePrivate()
 
     bufferRef = QWaylandBufferRef();
 
-    foreach (QtWayland::FrameCallback *c, pendingFrameCallbacks)
+    for (QtWayland::FrameCallback *c : qAsConst(pendingFrameCallbacks))
         c->destroy();
-    foreach (QtWayland::FrameCallback *c, frameCallbacks)
+    for (QtWayland::FrameCallback *c : qAsConst(frameCallbacks))
         c->destroy();
 }
 
-void QWaylandSurfacePrivate::setSize(const QSize &s)
-{
-    Q_Q(QWaylandSurface);
-    if (size != s) {
-        opaqueRegion = QRegion();
-        size = s;
-        q->sizeChanged();
-    }
-}
-
-void QWaylandSurfacePrivate::setBufferScale(int scale)
-{
-    Q_Q(QWaylandSurface);
-    if (scale == bufferScale)
-        return;
-    bufferScale = scale;
-    emit q->bufferScaleChanged();
-}
-
 void QWaylandSurfacePrivate::removeFrameCallback(QtWayland::FrameCallback *callback)
 {
     pendingFrameCallbacks.removeOne(callback);
@@ -195,7 +142,8 @@ void QWaylandSurfacePrivate::removeFrameCallback(QtWayland::FrameCallback *callb
 void QWaylandSurfacePrivate::notifyViewsAboutDestruction()
 {
     Q_Q(QWaylandSurface);
-    foreach (QWaylandView *view, views) {
+    const auto viewsCopy = views; // Views will be removed from the list when marked as destroyed
+    for (QWaylandView *view : viewsCopy) {
         QWaylandViewPrivate::get(view)->markSurfaceAsDestroyed(q);
     }
     if (hasContent) {
@@ -261,7 +209,7 @@ void QWaylandSurfacePrivate::surface_frame(Resource *resource, uint32_t callback
 
 void QWaylandSurfacePrivate::surface_set_opaque_region(Resource *, struct wl_resource *region)
 {
-    opaqueRegion = region ? QtWayland::Region::fromResource(region)->region() : QRegion();
+    pending.opaqueRegion = region ? QtWayland::Region::fromResource(region)->region() : QRegion();
 }
 
 void QWaylandSurfacePrivate::surface_set_input_region(Resource *, struct wl_resource *region)
@@ -277,42 +225,74 @@ void QWaylandSurfacePrivate::surface_commit(Resource *)
 {
     Q_Q(QWaylandSurface);
 
+    // Needed in order to know whether we want to emit signals later
+    QSize oldBufferSize = bufferSize;
+    QRectF oldSourceGeometry = sourceGeometry;
+    QSize oldDestinationSize = destinationSize;
+    bool oldHasContent = hasContent;
+    int oldBufferScale = bufferScale;
+
+    // Update all internal state
     if (pending.buffer.hasBuffer() || pending.newlyAttached)
         bufferRef = pending.buffer;
-
-    auto buffer = bufferRef.buffer();
-    if (buffer)
-        buffer->setCommitted(pending.damage);
-
-    setSize(bufferRef.size());
-    damage = pending.damage.intersected(QRect(QPoint(), size));
-
-    for (int i = 0; i < views.size(); i++) {
-        views.at(i)->bufferCommitted(bufferRef, damage);
-    }
-
-    emit q->damaged(damage);
-
-    bool oldHasContent = hasContent;
+    bufferScale = pending.bufferScale;
+    bufferSize = bufferRef.size();
+    QSize surfaceSize = bufferSize / bufferScale;
+    sourceGeometry = !pending.sourceGeometry.isValid() ? QRect(QPoint(), surfaceSize) : pending.sourceGeometry;
+    destinationSize = pending.destinationSize.isEmpty() ? sourceGeometry.size().toSize() : pending.destinationSize;
+    damage = pending.damage.intersected(QRect(QPoint(), destinationSize));
     hasContent = bufferRef.hasContent();
-    if (oldHasContent != hasContent)
-        emit q->hasContentChanged();
-
-    if (!pending.offset.isNull())
-        emit q->offsetForNextFrame(pending.offset);
-
-    setBufferScale(pending.bufferScale);
+    frameCallbacks << pendingFrameCallbacks;
+    inputRegion = pending.inputRegion.intersected(QRect(QPoint(), destinationSize));
+    opaqueRegion = pending.opaqueRegion.intersected(QRect(QPoint(), destinationSize));
+    QPoint offsetForNextFrame = pending.offset;
 
+    if (viewport)
+        viewport->checkCommittedState();
 
+    // Clear per-commit state
     pending.buffer = QWaylandBufferRef();
     pending.offset = QPoint();
     pending.newlyAttached = false;
     pending.damage = QRegion();
-
-    frameCallbacks << pendingFrameCallbacks;
     pendingFrameCallbacks.clear();
 
-    inputRegion = pending.inputRegion.intersected(QRect(QPoint(), size));
+    // Notify buffers and views
+    if (auto *buffer = bufferRef.buffer())
+        buffer->setCommitted(damage);
+    for (auto *view : qAsConst(views))
+        view->bufferCommitted(bufferRef, damage);
+
+    // Now all double-buffered state has been applied so it's safe to emit general signals
+    // i.e. we won't have inconsistensies such as mismatched surface size and buffer scale in
+    // signal handlers.
+
+    emit q->damaged(damage);
+
+    if (oldBufferSize != bufferSize) {
+        emit q->bufferSizeChanged();
+#if QT_DEPRECATED_SINCE(5, 13)
+QT_WARNING_PUSH
+QT_WARNING_DISABLE_DEPRECATED
+        emit q->sizeChanged();
+QT_WARNING_POP
+#endif
+    }
+
+    if (oldBufferScale != bufferScale)
+        emit q->bufferScaleChanged();
+
+    if (oldDestinationSize != destinationSize)
+        emit q->destinationSizeChanged();
+
+    if (oldSourceGeometry != sourceGeometry)
+        emit q->sourceGeometryChanged();
+
+    if (oldHasContent != hasContent)
+        emit q->hasContentChanged();
+
+    if (!offsetForNextFrame.isNull())
+        emit q->offsetForNextFrame(offsetForNextFrame);
 
     emit q->redraw();
 }
@@ -405,7 +385,8 @@ QWaylandSurface::QWaylandSurface(QWaylandSurfacePrivate &dptr)
 QWaylandSurface::~QWaylandSurface()
 {
     Q_D(QWaylandSurface);
-    QWaylandCompositorPrivate::get(d->compositor)->unregisterSurface(this);
+    if (d->compositor)
+        QWaylandCompositorPrivate::get(d->compositor)->unregisterSurface(this);
     d->notifyViewsAboutDestruction();
 }
 
@@ -459,11 +440,22 @@ QWaylandClient *QWaylandSurface::client() const
 {
     Q_D(const QWaylandSurface);
     if (isDestroyed() || !compositor() || !compositor()->clients().contains(d->client))
-        return Q_NULLPTR;
+        return nullptr;
 
     return d->client;
 }
 
+/*!
+ * Holds the \c wl_client using this QWaylandSurface.
+ */
+::wl_client *QWaylandSurface::waylandClient() const
+{
+    if (auto *c = client())
+        return c->client();
+
+    return nullptr;
+}
+
 /*!
  * \qmlproperty bool QtWaylandCompositor::WaylandSurface::hasContent
  *
@@ -482,21 +474,109 @@ bool QWaylandSurface::hasContent() const
 }
 
 /*!
- * \qmlproperty size QtWaylandCompositor::WaylandSurface::size
+ * \qmlproperty rect QtWaylandCompositor::WaylandSurface::sourceGeometry
+ * \since 5.13
  *
- * This property holds the WaylandSurface's size in pixels.
+ * This property describes the portion of the attached Wayland buffer that should
+ * be drawn on the screen. The coordinates are from the corner of the buffer and are
+ * scaled by \l bufferScale.
+ *
+ * \sa bufferScale
+ * \sa bufferSize
+ * \sa destinationSize
+ */
+
+/*!
+ * \property QWaylandSurface::sourceGeometry
+ * \since 5.13
+ *
+ * This property describes the portion of the attached QWaylandBuffer that should
+ * be drawn on the screen. The coordinates are from the corner of the buffer and are
+ * scaled by \l bufferScale.
+ *
+ * \sa bufferScale
+ * \sa bufferSize
+ * \sa destinationSize
+ */
+QRectF QWaylandSurface::sourceGeometry() const
+{
+    Q_D(const QWaylandSurface);
+    return d->sourceGeometry;
+}
+
+/*!
+ * \qmlproperty size QtWaylandCompositor::WaylandSurface::destinationSize
+ * \since 5.13
+ *
+ * This property holds the size of this WaylandSurface in surface coordinates.
+ *
+ * \sa bufferScale
+ * \sa bufferSize
+ */
+
+/*!
+ * \property QWaylandSurface::destinationSize
+ * \since 5.13
+ *
+ * This property holds the size of this WaylandSurface in surface coordinates.
+ *
+ * \sa bufferScale
+ * \sa bufferSize
+ */
+QSize QWaylandSurface::destinationSize() const
+{
+    Q_D(const QWaylandSurface);
+    return d->destinationSize;
+}
+
+/*!
+ * \qmlproperty size QtWaylandCompositor::WaylandSurface::bufferSize
+ *
+ * This property holds the size of the current buffer of this WaylandSurface in pixels,
+ * not in surface coordinates.
+ *
+ * For the size in surface coordinates, use \l destinationSize instead.
+ *
+ * \sa destinationSize
+ * \sa bufferScale
+ */
+
+/*!
+ * \property QWaylandSurface::bufferSize
+ *
+ * This property holds the size of the current buffer of this QWaylandSurface in pixels,
+ * not in surface coordinates.
+ *
+ * For the size in surface coordinates, use \l destinationSize instead.
+ *
+ * \sa destinationSize
+ * \sa bufferScale
+ */
+QSize QWaylandSurface::bufferSize() const
+{
+    Q_D(const QWaylandSurface);
+    return d->bufferSize;
+}
+
+#if QT_DEPRECATED_SINCE(5, 13)
+/*!
+ * \qmlproperty size QtWaylandCompositor::WaylandSurface::size
+ * \obsolete use bufferSize or destinationSize instead
+ *
+ * This property has been deprecated, use \l bufferSize or \l destinationSize instead.
  */
 
 /*!
  * \property QWaylandSurface::size
+ * \obsolete use bufferSize or destinationSize instead
  *
- * This property holds the QWaylandSurface's size in pixels.
+ * This property has been deprecated, use \l bufferSize or \l destinationSize instead.
  */
 QSize QWaylandSurface::size() const
 {
-    Q_D(const QWaylandSurface);
-    return d->size;
+    return bufferSize();
 }
+#endif
 
 /*!
  * \qmlproperty size QtWaylandCompositor::WaylandSurface::bufferScale
@@ -589,7 +669,7 @@ QWaylandCompositor *QWaylandSurface::compositor() const
 void QWaylandSurface::frameStarted()
 {
     Q_D(QWaylandSurface);
-    foreach (QtWayland::FrameCallback *c, d->frameCallbacks)
+    for (QtWayland::FrameCallback *c : qAsConst(d->frameCallbacks))
         c->canSend = true;
 }
 
@@ -603,7 +683,7 @@ void QWaylandSurface::sendFrameCallbacks()
     int i = 0;
     while (i < d->frameCallbacks.size()) {
         if (d->frameCallbacks.at(i)->canSend) {
-            d->frameCallbacks.at(i)->surface = Q_NULLPTR;
+            d->frameCallbacks.at(i)->surface = nullptr;
             d->frameCallbacks.at(i)->send(time);
             d->frameCallbacks.removeAt(i);
         } else {
@@ -622,6 +702,24 @@ bool QWaylandSurface::inputRegionContains(const QPoint &p) const
     return d->inputRegion.contains(p);
 }
 
+/*!
+ * Returns \c true if the QWaylandSurface's input region contains the point \a position.
+ * Otherwise returns \c false.
+ *
+ * \since 5.14
+ */
+bool QWaylandSurface::inputRegionContains(const QPointF &position) const
+{
+    Q_D(const QWaylandSurface);
+    // QRegion::contains operates in integers. If a region has a rect (0,0,10,10), (0,0) is
+    // inside while (10,10) is outside. Therefore, we can't use QPoint::toPoint(), which will
+    // round upwards, meaning the point (-0.25,-0.25) would be rounded to (0,0) and count as
+    // being inside the region, and similarly, a point (9.75,9.75) inside the region would be
+    // rounded upwards and count as being outside the region.
+    const QPoint floored(qFloor(position.x()), qFloor(position.y()));
+    return d->inputRegion.contains(floored);
+}
+
 /*!
  * \qmlmethod void QtWaylandCompositor::WaylandSurface::destroy()
  *
@@ -679,6 +777,31 @@ bool QWaylandSurface::isCursorSurface() const
     return d->isCursorSurface;
 }
 
+/*!
+ * \qmlproperty bool QtWaylandCompositor::WaylandSurface::inhibitsIdle
+ * \since 5.14
+ *
+ * This property holds whether this surface is intended to inhibit the idle
+ * behavior of the compositor such as screen blanking, locking and screen saving.
+ *
+ * \sa IdleInhibitManagerV1
+ */
+
+/*!
+ * \property QWaylandSurface::inhibitsIdle
+ * \since 5.14
+ *
+ * This property holds whether this surface is intended to inhibit the idle
+ * behavior of the compositor such as screen blanking, locking and screen saving.
+ *
+ * \sa QWaylandIdleInhibitManagerV1
+ */
+bool QWaylandSurface::inhibitsIdle() const
+{
+    Q_D(const QWaylandSurface);
+    return !d->idleInhibitors.isEmpty();
+}
+
 #if QT_CONFIG(im)
 QWaylandInputMethodControl *QWaylandSurface::inputMethodControl() const
 {
@@ -716,7 +839,7 @@ QWaylandView *QWaylandSurface::primaryView() const
 {
     Q_D(const QWaylandSurface);
     if (d->views.isEmpty())
-        return Q_NULLPTR;
+        return nullptr;
     return d->views.first();
 }
 
@@ -761,11 +884,13 @@ QList<QWaylandView *> QWaylandSurface::views() const
 }
 
 /*!
- * Returns the QWaylandSurface corresponding to the Wayland resource \a res.
+ * Returns the QWaylandSurface corresponding to the Wayland resource \a resource.
  */
-QWaylandSurface *QWaylandSurface::fromResource(::wl_resource *res)
+QWaylandSurface *QWaylandSurface::fromResource(::wl_resource *resource)
 {
-    return static_cast<QWaylandSurfacePrivate *>(QWaylandSurfacePrivate::Resource::fromResource(res)->surface_object)->q_func();
+    if (auto p = QtWayland::fromResource<QWaylandSurfacePrivate *>(resource))
+        return p->q_func();
+    return nullptr;
 }
 
 /*!
@@ -778,11 +903,12 @@ struct wl_resource *QWaylandSurface::resource() const
 }
 
 /*!
- * Sets a \a role on the surface. A role defines how a surface will be mapped on screen, without a role
- * a surface is supposed to be hidden. Only one role at all times can be set on a surface. Although
+ * Sets a \a role on the surface. A role defines how a surface will be mapped on screen; without a role
+ * a surface is supposed to be hidden. Only one role can be set on a surface, at all times. Although
  * setting the same role many times is allowed, attempting to change the role of a surface will trigger
  * a protocol error to the \a errorResource and send an \a errorCode to the client.
  *
+ * Returns true if a role can be assigned; false otherwise.
  */
 bool QWaylandSurface::setRole(QWaylandSurfaceRole *role, wl_resource *errorResource, uint32_t errorCode)
 {
@@ -808,7 +934,7 @@ QWaylandSurfaceRole *QWaylandSurface::role() const
 
 QWaylandSurfacePrivate *QWaylandSurfacePrivate::get(QWaylandSurface *surface)
 {
-    return surface ? surface->d_func() : Q_NULLPTR;
+    return surface ? surface->d_func() : nullptr;
 }
 
 void QWaylandSurfacePrivate::ref()
@@ -844,7 +970,7 @@ void QWaylandSurfacePrivate::derefView(QWaylandView *view)
 void QWaylandSurfacePrivate::initSubsurface(QWaylandSurface *parent, wl_client *client, int id, int version)
 {
     Q_Q(QWaylandSurface);
-    QWaylandSurface *oldParent = 0; // TODO: implement support for switching parents
+    QWaylandSurface *oldParent = nullptr; // TODO: implement support for switching parents
 
     subsurface = new Subsurface(this);
     subsurface->init(client, id, version);
@@ -887,6 +1013,30 @@ void QWaylandSurfacePrivate::Subsurface::subsurface_set_desync(wl_subsurface::Re
     qDebug() << Q_FUNC_INFO;
 }
 
+/*!
+ * \qmlsignal QtWaylandCompositor::WaylandSurface::childAdded(WaylandSurface child)
+ *
+ * This signal is emitted when a wl_subsurface, \a child, has been added to the surface.
+ */
+
+/*!
+ * \fn void QWaylandSurface::childAdded(QWaylandSurface *child)
+ *
+ * This signal is emitted when a wl_subsurface, \a child, has been added to the surface.
+ */
+
+/*!
+ * \qmlsignal QtWaylandCompositor::WaylandSurface::surfaceDestroyed()
+ *
+ * This signal is emitted when the corresponding wl_surface is destroyed.
+ */
+
+/*!
+ * \fn void QWaylandSurface::surfaceDestroyed()
+ *
+ * This signal is emitted when the corresponing wl_surface is destroyed.
+ */
+
 /*!
  * \qmlsignal void QtWaylandCompositor::WaylandSurface::dragStarted(WaylandDrag drag)
  *
@@ -899,4 +1049,21 @@ void QWaylandSurfacePrivate::Subsurface::subsurface_set_desync(wl_subsurface::Re
  * This signal is emitted when a \a drag has started from this surface.
  */
 
+/*!
+ * \fn void damaged(const QRegion &rect)
+ *
+ * This signal is emitted when the client tells the compositor that a particular part of, or
+ * possibly the entire surface has been updated, so the compositor can redraw that part.
+ *
+ * While the compositor APIs take care of redrawing automatically, this function may be useful
+ * if you require a specific, custom behavior.
+ */
+
+/*!
+ * \fn void parentChanged(QWaylandSurface *newParent, QWaylandSurface *oldParent)
+ *
+ * This signal is emitted when the client has requested that this surface should be a
+ * subsurface of \a newParent.
+ */
+
 QT_END_NAMESPACE
